////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
// 1.3.1 "project" operation

// Obtain the names of relation tuple's attributes
String[] attrNames = operand.getTupleType().getAttributeNames();
// Obtain the list of attrbutes to be removed 
String[] attrsToRemove = (String[]) parameter;

// If any of the elements of attrsToRemove cannot be found in the attrNames arrray,
// IAE should be thrown due to the wrong names of attribute to remove

String[] newAttrNames = // elements of attrNames array, except those which are contained in attrsToRemove array
 
// Obtain the types of attributes
Class[] newAttrTypes = new Class[newAttrNames.length];
for (int i = 0; i < newAttrNames.length; ++i)
	newAttrTypes[i] = operand.getTupleType().getAttributeType(newAttrNames[i]);

// Create a new TupleType
TupleType tupleType = new TupleType(newAttrNames, newAttrTypes);

// Create a new relation using RelationBuilder
Relation relation = builder.createRelation(tupleType);

for (Iterator iterator = operand.getTuples().iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	
	// Create a new tuple using RelationBuilder
	Tuple newTuple = builder.createTuple(tupleType);
	
	// Fill it with values
	for (int i = 0; i < newAttrNames.length; ++i)
		newTuple.setObject(newAttrNames[i], tuple.getObject(newAttrNames[i]));
		
	// Add it to the relation
	relation.addTuple(newTuple);
}

// Return result relation
return relation;



//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.2 "alias" operation

// Obtain the names of relation tuple's attributes
String[] attrNames = operand.getTupleType().getAttributeNames();
// Obtain the mappings of attrbutes to be renamed 
Map aliasMap  = (Map) parameter;

// If any of the keys of aliasMap cannot be found in the attrNames arrray,
// IAE should be thrown due to the wrong names of attributes to rename

String[] newAttrNames = // elements of attrNames array, renamed according to the mappings
// The keys of the map specify the names of attributes to be renamed,
// the appropriate values of the map specify the new names of attributes.
 
 
// Obtain the types of attributes
Class[] attrTypes = new Class[attrNames.length];
for (int i = 0; i < attrNames.length; ++i)
	attrTypes[i] = operand.getTupleType().getAttributeType(attrNames[i]);

// Create a new TupleType
TupleType tupleType = new TupleType(newAttrNames, attrTypes);


// Create a new relation using RelationBuilder
Relation relation = builder.createRelation(tupleType);

for (Iterator iterator = operand.getTuples().iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	
	// Create a new tuple using RelationBuilder
	Tuple newTuple = builder.createTuple(tupleType);
	
	// Fill it with values
	for (int i = 0; i < newAttrNames.length; ++i)
		newTuple.setObject(newAttrNames[i], tuple.getObject(attrNames[i]));
		
	// Add it to the relation
	relation.addTuple(newTuple);
}

// Return result relation
return relation;


////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.3 "restrict" operation


// Obtain the names of relation tuple's attributes
String[] attrNames = operand.getTupleType().getAttributeNames();

// Obtain the restrict condition expression
String condition = (String)(((Object[]) parameter)[0]);

// Obtain the additional variable mappings
Map paramVariables = (Map)(((Object[]) parameter)[1]);

// Create a HashMap with paramVariables contents copied to it
Map variables = new HashMap(paramVariables);

// Get the set of tuples
Set tuples = operand.getTuples();

// Obtain an instance of Complex Boolean Evaluator
Evaluator evaluator = new EvaluatorFactory().getEvaluator(); 


// Create the result relation using RelationBuilder
Relation relation = builder.createRelation(operand.getTupleType());

// Iterate through all the tuples and remove those, for which condition is not true
for (Iterator iterator = tuples.iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	
	// Update the variable mappings to reflect the current tuple
	for (int i = 0; i < attrNames.length; ++i)
		variables.put(attrNames[i], tuple.getObject(i));

	// Evaluate condition expression
	boolean isTrue = evaluator.eval(condition, variables);
	
	// If condition is satisfied, add current tuple to the result relation
	if (isTrue)
		relation.addTuple(tuple);
		
}

// Return result relation
return relation;



///////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.4 "union" operation

if (!leftOperand.getTupleType().equals(rightOperand.getTupleType())) {
	// Tuple types of relations doesn't match, should throw an IAE
} 

// Create a result relation
Relation relation = leftOperand.getBuilder().createRelation(resultTuples);

Set leftTuples = leftOperand.getTuples();
Set rightTuples = rightOperand.getTuples();

// Add all the tuples from leftOperand to the relation
for (Iterator iterator = leftTuples.iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	relation.addTuple(tuple);
}
// Add all the tuples from leftOperand to the relation (similar to the above)

// Return the result
return relation;


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.5 "intersect" operation


if (!leftOperand.getTupleType().equals(rightOperand.getTupleType())) {
	// Tuple types of relations doesn't match, should throw an IAE
} 

// Create a result relation
Relation relation = leftOperand.getBuilder().createRelation(resultTuples);

Set leftTuples = leftOperand.getTuples();
Set rightTuples = rightOperand.getTuples();

// Loop through all the tuples from leftOperand
for (Iterator iterator = leftTuples.iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	// If the rightOperand also contains this tuple, add it to result relation
	if (rightTuples.contains(tuple))
		relation.addTuple(tuple);
}
// Return the result
return relation;

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.6 "difference" operation

if (!leftOperand.getTupleType().equals(rightOperand.getTupleType())) {
	// Tuple types of relations doesn't match, should throw an IAE
} 

// Create a result relation
Relation relation = leftOperand.getBuilder().createRelation(resultTuples);

Set leftTuples = leftOperand.getTuples();
Set rightTuples = rightOperand.getTuples();

// Loop through all the tuples from leftOperand
for (Iterator iterator = leftTuples.iterator(); iterator.hasNext(); ) {
	Tuple tuple = (Tuple) iterator.next();
	// If the rightOperand doesn't contain this tuple, add it to result relation
	if (!rightTuples.contains(tuple))
		relation.addTuple(tuple);
}
// Return the result
return relation;

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
// 1.3.7 Perform "product" operation

// Obtain the names of tuple attributes for both relations
String[] leftOperandAttrs = leftOperand.getTupleType().getAttributeNames();
String[] rightOperandAttrs.getTupleType().getAttributesNames();

// Check if there are no attributes with the same names
for (int i = 0; i < leftOperandAttrs.length; ++i) 
	for (int j = 0; j < rightOperandAttrs.length; ++j) 
		if (leftOperandAttrs[i].equals(rightOperandAttrs[j]))
			// throw IAE here, because of equal attribute names			

// Get the types of tuple attributes of the left operand
Class[] leftOpAttrTypes = new Class [leftOperandAttrs.length];
for (int i = 0; i < leftOpAttrTypes.length; ++i)
	leftOpAttrTypes[i] =  leftOperand.getTupleType().getAttributeType(leftOperandAttr[i]);
// Get the types of tuple attributes of the right operand, similar to the above code,
// store them in rightOpAttrTypes array

// Concatenate the obtained arrays
String[] newAttrs = // concatenation of leftOperandAttr and rightOperandAttr arrays
Class[] newAttrTypes = // concatenation of leftOpAttrTypes and rightOpAttrTypes arrays
	
// Create a new tuple type
TupleType tupleType = new TupleType(newAttrs, newAttrTypes);

// Create a new relation using RelationBuilder
Relation relation = builder.createRelation(tupleType);

// Perform the product operation itself
for (Iterator leftIter = leftOperand.getTuples().iterator(); leftIter.hasNext(); ) {
	Tuple leftTuple = (Tuple) leftIter.next();
	
	for (Iterator rightIter = rightOperand.getTuples().iterator(); rightIter.hasNext(); ) {
		Tuple rightTuple = (Tuple) rightIter.next();
		
		// Create a new tuple using RelationBuilder
		Tuple newTuple = builder.createTuple(tupleType);
		
		// Fill it with values
		for (int i = 0; i < leftOperandAttr.length; ++i)
			newTuple.setObject(leftOperandAttr[i], leftTuple.getObject(leftOperandAttr[i]));
		for (int i = 0; i < rightOperandAttr.length; ++i)
			newTuple.setObject(rightOperandAttr[i], rightTuple.getObject(rightOperandAttr[i]));
		
		// Add it to the relation
		relation.addTuple(newTuple);
	} 
}

// Return the result relation
return relation;

	

///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
// 1.3.8 Perform "conditional join" operation


// Obtain the names of tuple attributes for both relations (similar to the like-named step of "product" operation)
// They should be stored in the leftOperandAttr and rightOperandAttr arrays
// Check if there are no attributes with the same names (similar to the like-named step of "product" operation)


// Get the list of conditions
JoinCondition[] conditions = (JoinCondition[]) parameter.clone();

// Attrbutes compare by the different sides of the condition should be of the same type
// If this is not met, throw IAE

// If the parameter is null, or conditions array has zero-length,
// the "product" operation should be performed, see above algorithm

// Then the list of conditions should be sorted in such a way,
// that more specific conditions are in the begin of the list and the less specific in the end
// The suggested way of doing it, is sorting conditions by the used operation
// Conditions with the "=" equality operation will go in the begin of the list
// Conditions with the greater/less/greater or equals/less or equal will go after them
// Note, that the correctness of result doesn't depend on the order of condition,
// but the performance of operation does depend

// The following steps are similar to the like-named steps of "product" operation)
// Get the types of tuple attributes of the left operand, store them in rightOpAttrTypes array 
// Get the types of tuple attributes of the right operand, store them in rightOpAttrTypes array  
// Concatenate the obtained arrays
String[] newAttrs = // concatenation of leftOperandAttr and rightOperandAttr arrays
Class[] newAttrTypes = // concatenation of leftOpAttrTypes and rightOpAttrTypes arrays
	
// Create a new tuple type, store it in tupleType variable
// Create a new relation using RelationBuilder, store it in the relation variable

Set resultTuples = new HashSet();
Set tmpTuples = new HashSet();

// This Map will store all the required indexes
Map indexes = new HashMap();

// Loop through all tuples contained in the leftOperand
for (Iterator iterator = leftOperand.getTuples().iterator(); iterator.hasNext(); ) {
	Tuple leftTuple = (Tuple) iterator.next();
	
	resultTuples.clear();
	
	// Loop through all the conditions and apply them
	for (int i = 0; i < conditions.length; ++i) {
		tmpTuples.clear();
	
		// Obtain index for the right attribute
		SortedMultiMap rightIndex = (SortedMultiMap)indexes.get(conditions[i].getRightAttributeName());
		if (rightIndex == null) {
			rightIndex = leftOperand.getIndex(conditions[i].getRightAttributeName());
			indexes.put(rightIndex);
		}	
		
		// Process the condition
		if (conditions[i].getOperation().equals("=") || conditions[i].getOperation().equals("<=") || 
						conditions[i].getOperation().equals(">=")) { 
			tmpTuples.addAll(rightIndex.getValues(leftTuple.getObject(conditions[i].getLeftAttributeName()))); 
		} else if (conditions[i].getOperation().equals("<") || conditions[i].getOperation().equals("<=")) {
			SortedMultiMap tail = rightIntex.getTailMultiMap(leftTuple.getObject(conditions[i].getLeftAttributeName()));
			Iterator valIt = tail.values().iterator();
			while (valIt.hasNext()) {
				tmpTuples.add(valIt.next());
			}  
		} else if (conditions[i].getOperation().equals(">") || conditions[i].getOperation().equals(">=")) {
			// The same as above, but the head MultiMap is obtained instead of tail
			// i.e. getHeadMultiMap is used
		}
		
		if (i = 0) {
			resultTuples.putAll(tmpTuples);
		} else {
			resultTuples.retainAll(tmpTuples);
		}
	}
	
	for (Iterator resIt = resultTuples.iterator(); resIt.hasNext(); ) {
		Tuple rightTuple = (Tuple) resIt.next();
	
		// Create a new tuple using RelationBuilder
		Tuple newTuple = builder.createTuple(tupleType);
		
		// Fill it with values
		for (int i = 0; i < leftOperandAttr.length; ++i)
			newTuple.setObject(leftOperandAttr[i], leftTuple.getObject(leftOperandAttr[i]));
		for (int i = 0; i < rightOperandAttr.length; ++i)
			newTuple.setObject(rightOperandAttr[i], rightTuple.getObject(rightOperandAttr[i]));
		
		// Add it to the relation
		relation.addTuple(newTuple);
	}
} 



///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
// 1.3.9 Perform "natural join" operation

The naturaml join operation should be perfromed in the way quite similar to the conditional join.
But there are some differences that should be payed attention to:
  - the join conditions are not specified, developer should assume 
that there are equality conditions between the same-named atttributes in the two relations
  - note, that having the same-named attributes is permitted, 
the result relation will contain only a single copy of such attributes 
Note, that if there are no same-named columns, the "product" operation will be performed

///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
// 1.3.10 Load the relation definition from configuration file and create relation


// Note, that this algorithm assumes that the configuration has valid format
// Actually the devloper should validate the configuration basing on the sample configuration file 

// Get a ConfigManager instance
ConfigManager manager = ConfigManager.getInstance();

// Get the property describing the relation
Property prop = manager.getProperty(manager.getPropertyObject(namespace, "Relations." + relationName)); 

// Get the "Attributes" nested property
Property attrProp = prop.getProperty("Attributes");

// Build the TupleType from the "Attributes" property
String[] attrNames = new String[attrProp.list().size()];
Class[] attrTypes = new Class[attrProp.list().size()];
int i = 0;
for (Iterator iterator = attrProp.list().iterator(); iterator.hasNext(); ++i) {
	Property subProp = (Property)iterator.next();
	attrNames[i] = subProp.getName();
	attrTypes[i] = Class.forName(attrTypes[i]);
}
TupleType tupleType = new TupleType(attrNames, attrTypes);

// Create a relation
Relation relation = createRelation(tupleType);

// Retrieve information about the tuple-level constraints, create them and add to the relation
Property constrProp = prop.getProperty("TupleConstraints");
int consCount = constrProp.list().size();
for (int i = 1; i <= consCount; ++i) {
	Property singleConsProp = constrProp.getProperty("" + i);
	// Obtain the RelationConstraint interface implementation class
	Class consClass = Class.forName(singleConsProp.getValue("Class")); 
	// Obtain the list of parameters to be passed to constructor
	Property paramProp = singleConsProp.getProperty("Parameters");
	String[] parameters = new String[paramProp.list().size()];
	for (Iterator paramIt = parapProp.list().iterator(); paramIt.hasNext(); ) {
		Property singleParamProp =  (Property) paramIt.next();
		parameters[Integer.parseInt(singleParamProp.getValue()) - 1] = singlParamProp.getValue();
	}
	Class[] paramTypes = new Class[];
	System.fill(paramTypes, String.class);
	// Create an instance of the constraint
	TupleConstraint cons = (TupleConstraint) consClass.getConstructor(paramTypes).newInstance(parameters); 
	// Add the constraint to the relation
	relation.addTupleConstraint(cons);
}
// Retrieve information about the relation-level constraints, create them and add to the relation (similar to above)

